



def create(iw,ih,c):
    '''crea una immagine ovvero una lista di liste del colore c 
    larga iw e alta ih'''
    img = [ [ c for _ in range(iw) ] for _ in range(ih) ]
    return img

##############################################à
    

def draw_rettangolo(img,x,y,w,h,c):
    '''disegna sulla immagine img alle coordinate x,y 
    un rettangolo di colore c largo w ed alto h'''
    ampiezza=len(img[0])
    altezza=len(img)
    for j in range(y,y+h):
        for i in range(x,x+w):
            if 0<=j<altezza and 0<=i<ampiezza:
                img[j][i] = c

###############################################
                
def draw_scacchiera(img,s,c0,c1):
    '''disegna sulla immagine img una scacchiera con quadrati di 
    lato s e colori c0 e 01'''
    h, w = len(img), len(img[0])
    for jj in range(h//s + 1): #righe della scacchiera
        for ii in range(w//s + 1): #colonne della scacchiera
            if (ii + jj) % 2: c=c1
            else: c=c0
            draw_rettangolo(img,ii*s,jj*s,s,s,c)

###############################################
            
def inverti(img):
    '''crea l'immagine negativa di img'''
    h, w = len(img), len(img[0])
    ret = create(w,h,(0,0,0))
    for j in range(h):
        for i in range(w):
            r, g, b = img[j][i]
            ret[j][i] = ( 255 - r, 255 - g, 255 - b)
    return ret

################################################
    
def mosaico(img, s):
    '''Ritorna una nuova immagine ottenuta dividendo l'immagine img 
    in quadratini di lato s e riempendo ogni 
    quadratino con il colore dello spigolo in alto a sinistra 
    del quadratino di img'''
    w, h = len(img[0]), len(img)
    ret = create(w, h, (0,0,0))
    for y in range(h//s):		# divisione intera
        for x in range(w//s):	# divisione intera
            c = img[y*s][x*s]	# colore dello spigolo in alto-sinistra 
            draw_rettangolo(ret, x*s, y*s, s, s, c)
    return ret


###############################################    
def mosaico_media(img, s):
    '''Ritorna una nuova immagine ottenuta dividendo l'immagine img in
    quadratini di lato s e riempendo ogni quadratino con la media dei
    colori dei pixels nel quadratino di img'''
    w, h = len(img[0]), len(img)
    ret = create(w, h, (0,0,0))
    for jj in range(h//s):
        for ii in range(w//s):
            #calcola il color medio
            c = [0,0,0]
            n = 0
            for j in range(jj*s, (jj+1)*s):            
                for i in range(ii*s, (ii+1)*s):
                    if j<len(img) and i<len(img[0]):	
                        n += 1
                        for k in range(3):          
                            c[k] += img[j][i][k]   
            for k in range(3):
                X = round(c[k] / n)			# media
                X = min(X, 255)				# entro 0-255
                X = max(X, 0)
                c[k] = X
            c=tuple(c)
            draw_rettangolo(ret, ii*s, jj*s, s, s, c)
    return ret

#################################################

def mosaico_size(img, s):
    '''Ritorna una nuova immagine ottenuta dividendo l'immagine img 
    in quadratini di lato s e disegnando all'interno di ognuno di essi, 
    su sfondo nero, un quadratino centrale bianco di lato proporzionale 
    alla luminosita' media del corrispondente quadratino di img'''
    w, h = len(img[0]), len(img)
    ret = create(w, h, (0,0,0))
    for jj in range(h//s):
        for ii in range(w//s):
            r = 0
            n = 0
            for j in range(jj*s, (jj+1)*s):	# per ogni pixel nel quadratino
                for i in range(ii*s, (ii+1)*s):
                    if i<len(img[0]) and j<len(img):
                        n += 1
                        for k in range(3):	# somma i valori di tutte le
                            r += img[j][i][k]	# componenti colore
            lm = r / (n*3*255)				# luminosita' media quadratino
            t = int(round(s * lm))			# lato del quadratino bianco
            draw_rettangolo(ret, ii*s+(s-t)//2, jj*s+(s-t)//2, t, t, (255,255,255))
    return ret


################################################
            
def draw_gradiente_orizzontale(img,c0,c1):
    '''colora l'immagine img con un gradiente orizzontale che va dal colore 
    c0 al colore c1'''
    h, w = len(img), len(img[0])
    r0, g0, b0 = c0
    r1, g1, b1 = c1
    for i in range(w):
        u = i / w
        r = round(r0 * (1-u) + r1 * u)
        g = round(g0 * (1-u) + g1 * u)
        b = round(b0 * (1-u) + b1 * u)
        for j in range(h):
            img[j][i] = (r,g,b)
